from .otBase import BaseTTXConverter


class table_G_P_O_S_(BaseTTXConverter):
    """Glyph Positioning table

    The ``GPOS`` table stores advanced glyph-positioning data
    used in OpenType Layout features, such as mark attachment,
    cursive attachment, kerning, and other position adjustments.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/gpos
    """

    pass
